
from rgt.AnnotationSet import AnnotationSet

gene_source = "/home/egg/Projects/SVN/reg-gen/data/hg19/teste.gtf"
#gene_source = "/home/egg/Projects/SVN/reg-gen/data/hg19/gencode_annotation.gtf"

tf_source = "/home/egg/Projects/SVN/reg-gen/data/motifs/test.mtf"

ann_set = AnnotationSet(gene_source,tf_source)



